"""
Arc radon forward and inverse transform
Copyright (C) 2016  Syed Tabish(syedtabish.a@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""
from libc.math cimport sin, cos, sqrt, acos, atan2, floor
import numpy as np
cimport numpy as np
cimport cython as cy
import scipy
import scipy.linalg as lin
import time
from matplotlib import pyplot as plt

DEF eps = 0.0000001
DEF PI = 3.14159

@cy.boundscheck(False)
cdef inline double sum_arc(double[:,:] cimage,long[:] r, long[:] c):
    cdef:
        Py_ssize_t i = 0
        Py_ssize_t l = r.shape[0]
        Py_ssize_t r_max = cimage.shape[0]
        Py_ssize_t c_max = cimage.shape[1]
        double sum_ = 0
    for i in range(l):
        if ( (r[i] >= r_max) or (r[i] < 0) or (c[i] >= c_max) or (c[i] < 0)):
            continue
        sum_ = sum_ + cimage[r[i], c[i]]
    return sum_

@cy.boundscheck(False)
@cy.cdivision(True)
@cy.initializedcheck(False)
cdef inline double[:,:]  inverse_truncated_array(double[:] s, int r):
    cdef:
        double[:,:] inverse = np.zeros((s.shape[0], s.shape[0]))
        Py_ssize_t i = 0
    for i in xrange(r):
        if s[i] > 0: 
            inverse[i,i] = 1./s[i]
    return inverse

@cy.boundscheck(False)
@cy.cdivision(True)
cdef double[:,:] compute_A(int n, int m, double R, double alpha, int param = 1000):
    cdef:
        Py_ssize_t i,k
        double h = R / m
        double[:,:] A_n = np.zeros((m,m))
    
    for i in xrange(m):
        #hyperbolic limit
        l = floor(R - sqrt(R**2 + (h*i)**2 - 2*(i*h)*R*cos(alpha)) ) 
        #linear limit
        #l = floor( (h*i) - 2*(h*i)*sin(alpha/2) )
        #full lower triangular
        #l = 0#uncomment for full circular(full matrix) inversion
        l = int(l/(h) )
        l = max(0,l)
        decay = 1 + 1./param
        for k in xrange(l,i+1):
            A_n[i,k] = a(i-k) * sqrt(h) * k_(i*h + eps, k*h + eps, n, R)
        for k in range(l):
            if (k < i):
                A_n[i,k] = (a(i-k) * sqrt(h) * k_(i*h + eps, k*h + eps, n, R))*(decay**(-(k - (l)+ 1)**2))
    
    return A_n


cdef inline double a(int i) nogil:
    if i == 0:
        #return 4./3 #computed value at diagonal
        return 0.1/3 #somehow this suppresses streak artifacts
	#possible reason: this is equvivalent to smoothing at the lower limit.
    else:
        return ((4./3) * ((i + 1)**1.5 - (2 * (i**1.5)) + (i - 1)**1.5))

cdef inline double _interpolate(double[:,:] polar_image, float r, float theta, int mode ):
    cdef:
        Py_ssize_t r_1 = int(r)
        Py_ssize_t th_1 = int(theta)
        Py_ssize_t r_2 = r_1 + 1
        Py_ssize_t th_2 = th_1 + 1
        double val = 0
        double x,x_11,x_12,x_21,x_22
        double y,y_11,y_12,y_21,y_22

    if(mode == 0):
        val = polar_image[r_1, th_1]
    elif(mode == 1):
        x = r*cos(theta)
        y = r*sin(theta)
        x_11 = r_1*cos(th_1)
        x_12 = (r_1 + 1)*cos(th_1) 
        x_21 = (r_1)*cos(th_2)
        x_22 = (r_1 + 1)*cos(th_2)
        y_11 = r_1*sin(th_1)
        y_12 = (r_1 + 1)*sin(th_1) 
        y_21 = (r_1)*sin(th_2)
        y_22 = (r_1 + 1)*sin(th_2)
        
        a = sqrt((x - x_11)**2 + (y - y_11)**2)
        b = sqrt((x - x_12)**2 + (y - y_12)**2)
        c = sqrt((x - x_21)**2 + (y - y_21)**2)
        d = sqrt((x - x_22)**2 + (y - y_22)**2)
        try: 
            val = ((0 + b + c + d)*polar_image[r_1, th_1] +\
                   (a + 0 + c + d)*polar_image[r_2, th_1] +\
                   (a + b + 0 + d)*polar_image[r_1, th_2] +\
                   (a + b + c + 0)*polar_image[r_2, th_2]) /\
                   (3*(a + b + c + d)) 
        except IndexError:
            val = polar_image[r_1, th_1]
    return val

def pol2cart(double[:,:] polar_image):
    cdef:
        Py_ssize_t i, j
        Py_ssize_t R = polar_image.shape[0]
        Py_ssize_t Theta = polar_image.shape[1]
        float r, th
        float c = R + 0.0
        
    cdef:
        double[:,:] cart_image = np.zeros((2*R, 2*R))

    for i in xrange(2*R):
        for j in xrange(2*R):
            r = sqrt((i - c)**2 + (j - c)**2)
            th = atan2((c - i), (j-c))
            if th < 0:
                th = (2*PI) + th
            th = Theta*(th/(2*PI))
            if((r < R)  and (th < Theta)):
                cart_image[i,j] = _interpolate(polar_image, r, th, 1)
            else:
                pass
    return np.asarray(cart_image) 

@cy.boundscheck(False)
@cy.cdivision(True)
cdef double K_(double rho, double u, int n, double R) nogil:
    cdef double value = 0
    cdef double numerator, denominator, arg
    arg = (((R-u)**2 + R**2 - rho**2)/(2*R*(R-u)))
    arg = min( max(-1.0, arg), 1.0)
    numerator = 4. * rho * (R - u) * cos(n* acos(arg))
    denominator =  sqrt((rho + u) * (2*R + rho - u) * (2*R - rho - u))
    return (numerator/denominator)

@cy.boundscheck(False)
@cy.cdivision(True)
cdef double  k_(double rho, double u, int n, double R) nogil:
    return K_(rho, u, n, R)/K_(rho, rho, n , R)

@cy.boundscheck(False)
@cy.cdivision(True)
cdef complex[:,:] ghat_(complex[:,:] g, double R):
    cdef:
        Py_ssize_t n,rho
        int M = g.shape[0]
        int N = g.shape[1]
        double h = (R - eps) / (M  + 0.0)
        complex[:,:] gh = np.empty((M,N)) + 0j
    for n in xrange(N):
        for rho in xrange(M):
            gh[rho, n] = g[rho, n] / K_(rho*h + eps, rho*h + eps, n, R)
    return gh

@cy.initializedcheck(False)
cdef double[:,:] inverse_tsvd(double[:,:] A, int r):
    U, s, Vt = lin.svd(A, full_matrices=False)
    cdef:
        double[:,:] sinv
        double[:,:] Ar_inv = np.zeros((A.shape[0],A.shape[1]))
        double[:,:] Ar_invu = np.zeros((A.shape[0],A.shape[1]))
    sinv = inverse_truncated_array(s, r)
    Ar_inv = np.dot(np.dot(Vt.T, sinv), U.T) 
    return Ar_inv

def inverse_radon(complex[:,:] g, double R, double alpha, int rank, int param = 1000):
    cdef:
        int M = g.shape[0]#rho_count
        int N = g.shape[1]#phi_count
        complex[:,:] F = np.empty((g.shape[0],1)) + 0j
    ghat = ghat_(g,R)
    alpha = PI*(alpha/180.0)
    for n in xrange(N):
        A_n = compute_A(n, M, R, alpha, param)
        A_n_inv  = inverse_tsvd(A_n,rank)
        ff = np.dot(A_n_inv, ghat[:, n])
        F = np.hstack((F,ff[:,np.newaxis]))
    F = F[:,1:]
    f = F[::-1,:]#f(R-u) = F(u), u_max-->rho_max-->R
    h1 = np.asarray(f[:,:(N/2)+1])
    h2 = np.asarray(f[:,1:(N/2)])
    ff = np.hstack((h1, h2.conjugate()[:,::-1]))
    return ff

def arc_radonv(const double[:,:] cimage, int alpha, int phi_max, int rho_max, int R):
    """radon transform with visialization
    """
    cdef:
        Py_ssize_t i = 0
        Py_ssize_t j = 0
        Py_ssize_t k = 0
        Py_ssize_t rows = cimage.shape[0]
        Py_ssize_t cols = cimage.shape[1]
        double span = (alpha/180.)*PI
    
    cdef:
        double[:,::cy.view.contiguous] radon = np.empty((rho_max,phi_max))
        double[::cy.view.contiguous] rho = np.linspace(1, R-1, rho_max, endpoint = False)
        double[::cy.view.contiguous] phi = np.linspace(0, 2*PI, phi_max)
        double[::cy.view.contiguous] angle_val = np.linspace(-span, span, 2*alpha)
        double[::cy.view.contiguous] angles = np.empty(2*alpha)
        long[::cy.view.contiguous] r = np.empty(2*alpha, dtype=np.int)
        long[::cy.view.contiguous] c = np.empty(2*alpha, dtype=np.int)
    import time 
    fig = plt.figure()
    ax1, ax2 = fig.add_subplot(121), fig.add_subplot(122)
    plt.ion()
    ax1, ax2 = ax1.imshow(np.asarray(cimage)), ax2.imshow(np.asarray(cimage))
    plt.show()
    for i in xrange(phi_max):
        for k in xrange(2*alpha):
            angles[k] = angle_val[k] - PI + phi[i]
        for j in xrange(rho_max):
            print 'rho value --> ' + str(j)
            test = np.copy(np.asarray(cimage))#np.zeros((rows,cols))
            for k in xrange(2*alpha):
                r[k] = int(rows/2 - R*sin(phi[i]) - rho[j]*sin(angles[k]))
                c[k] = int(cols/2 + R*cos(phi[i]) + rho[j]*cos(angles[k]))
            radon[j,i] = sum_arc(cimage,r,c)
            for p in xrange(2*alpha):
                test[ r[p]-1:r[p]+1, c[p]-1:c[p]+1 ] = radon[j,i]  + 1
            ax1.set_data(test)
            ax2.set_data(radon)
            fig.canvas.draw()
            time.sleep(0.02)
    return np.asarray(radon)

@cy.boundscheck(False)
def arc_radon(const double[:,:] cimage, int alpha, int phi_max, int rho_max, int R):
    """circular arc radon transform
    """
    cdef:
        Py_ssize_t i = 0
        Py_ssize_t j = 0
        Py_ssize_t k = 0
        Py_ssize_t rows = cimage.shape[0]
        Py_ssize_t cols = cimage.shape[1]
        double span = (alpha/180.)*PI
    
    cdef:
        double[:,::cy.view.contiguous] radon = np.empty((rho_max,phi_max))
        double[::cy.view.contiguous] rho = np.linspace(1, R-1, rho_max, endpoint = False)
        double[::cy.view.contiguous] phi = np.linspace(0, 2*PI, phi_max)
        double[::cy.view.contiguous] RCphi = R*np.cos(phi)
        double[::cy.view.contiguous] RSphi = R*np.sin(phi)
        double[::cy.view.contiguous] angle_val = np.linspace(-span, span, 2*alpha)
        double[::cy.view.contiguous] angles = np.empty(2*alpha)
        long[::cy.view.contiguous] r = np.empty(2*alpha, dtype=np.int)
        long[::cy.view.contiguous] c = np.empty(2*alpha, dtype=np.int)
    for i in xrange(phi_max):
        for k in xrange(2*alpha):
            angles[k] = angle_val[k] - PI + phi[i]
        for j in xrange(rho_max):
            for k in xrange(2*alpha):
                r[k] = int(rows/2 - RSphi[i] - rho[j]*sin(angles[k]))
                c[k] = int(cols/2 + RCphi[i] + rho[j]*cos(angles[k]))
            radon[j,i] = sum_arc(cimage,r,c)
    return np.asarray(radon)

@cy.boundscheck(False)
def adjoint(const double[:,:] g, int R):
    """Adjoint based inversion
    """
    cdef:
        Py_ssize_t r = g.shape[0]*2
        Py_ssize_t c = g.shape[0]*2
        Py_ssize_t phi_max = g.shape[1]
        double sum_
        int rho 
        Py_ssize_t i, j, x, y

    cdef:
        double[:,::cy.view.contiguous] adjoint = np.empty((r,c))
        double[::cy.view.contiguous] RCphi =  R*np.cos(np.linspace(2*PI,0, phi_max))
        double[::cy.view.contiguous] RSphi =  R*np.sin(np.linspace(2*PI,0, phi_max))
    
    for i in range(r):
        for j in range(c):
            sum_ = 0
            x = j - r/2
            y = i - c/2
            for k in range(phi_max):
                rho = int(sqrt((x - RCphi[k])**2 + (y - RSphi[k])**2 ))
                if(rho < r/2):
                    sum_ = sum_ + g[rho, k]
            adjoint[i,j] = sum_
    return np.asarray(adjoint)

cdef double[:,:] compute_A_ext(int n, int m, double R, double alpha, int param = 1000):
    cdef:
        Py_ssize_t i,k
        double h = R / m
        double[:,:] A_n = np.zeros((m,m))
    
    for i in xrange(m):
        #hyperbolic limit
        l = floor(sqrt(R**2 + (h*i)**2 + 2*(i*h)*R*cos(alpha)) - R) 
        l = int(l/(h) )
        l = max(0,l)
        #l = 0
        l1 = 0
        decay = 1 + 1./param
        for k in xrange(l,i+1):
            A_n[i,k] = a(i-k) * sqrt(h) * k_(i*h + eps, k*h + eps, n, R)
        for k in range(l1, l):
            if (k < i):
                A_n[i,k] = (a(i-k) * sqrt(h) * k_(i*h + eps, k*h + eps, n, R))*(decay**(-(k - (l)+ 1)**2))
    
    return A_n

@cy.boundscheck(False)
def inverse_radon_ext(complex[:,:] g, double R, double alpha, int rank, int param = 1000):
    cdef:
        int M = g.shape[0]#rho_count
        int N = g.shape[1]#phi_count
        complex[:,:] F = np.empty((g.shape[0],1)) + 0j
        complex[:,:] zrs = np.zeros((R,N)) + 0j
    ghat = ghat_(g,R)
    alpha = np.pi*(alpha/180.0)
    np.set_printoptions(precision = 3, suppress=True)
    for n in xrange(N):
        A_n = compute_A_ext(n, M, R, alpha,param)
        A_n_inv  = inverse_tsvd(A_n,rank)
        ff = np.dot(A_n_inv, ghat[:, n])
        F = np.hstack((F,ff[:,np.newaxis]))
    F = F[:,1:]
    f = np.vstack((zrs,F));
    h1 = np.asarray(f[:,:(N/2)+1])
    h2 = np.asarray(f[:,1:(N/2)])
    ff = np.hstack((h1, h2.conjugate()[:,::-1]))
    return ff


def arc_radon_ext(const double[:,:] cimage, int alpha, int phi_max, int rho_max, int R):
    cdef:
        Py_ssize_t i = 0
        Py_ssize_t j = 0
        Py_ssize_t k = 0
        Py_ssize_t rows = cimage.shape[0]
        Py_ssize_t cols = cimage.shape[1]
        #Py_ssize_t R = min(rows, cols)//2
        float span = (alpha/180.)*PI
    
    cdef:
        double[:,::cy.view.contiguous] radon = np.empty((rho_max,phi_max))
        double[::cy.view.contiguous] rho = np.linspace(1, 2*R-1, rho_max, endpoint = False)
        double[::cy.view.contiguous] phi = np.linspace(0, 2*np.pi, phi_max)
        double[::cy.view.contiguous] RCphi = R*np.cos(phi)
        double[::cy.view.contiguous] RSphi = R*np.sin(phi)
        double[::cy.view.contiguous] angle_val = np.linspace(-span, span, 2*alpha)
        double[::cy.view.contiguous] angles = np.empty(2*alpha)
        long[::cy.view.contiguous] r = np.empty(2*alpha, dtype=np.int)
        long[::cy.view.contiguous] c = np.empty(2*alpha, dtype=np.int)
    for i in xrange(phi_max):
        for k in xrange(2*alpha):
            angles[k] = angle_val[k] + phi[i]
        for j in xrange(rho_max):
            for k in xrange(2*alpha):
                r[k] = int(rows/2 - RSphi[i] - rho[j]*sin(angles[k]))
                c[k] = int(cols/2 + RCphi[i] + rho[j]*cos(angles[k]))
            radon[j,i] = sum_arc(cimage,r,c)# / (2*alpha)
    return np.asarray(radon)
